from enum import Enum

from zbxtemplar.entities import Item
from zbxtemplar.entities.Dashboard import Widget, WidgetField, WidgetFieldType

class ItemHistoryHeader(str, Enum):
    NO = 0
    HORIZONTAL = 1
    VERTICAL = 2

class ItemHistory(Widget):
    def __init__(self, x: int = 0, y: int = 0, width: int = 12, height: int = 5,
                 name: str = ""):
        super().__init__(x, y, width, height, name)
        self._item_count = 0

    @property
    def type(self) -> str:
        return "itemhistory"

    def widget_fields(self) -> list:
        return self.fields

    def add_item(self, item: Item, name: str = ""):
        prefix = f"columns.{self._item_count}."
        self.fields.append(WidgetField(WidgetFieldType.ITEM, f"{prefix}itemid", {"host": item._host, "key": item.key}))
        self.fields.append(WidgetField(WidgetFieldType.STRING, f"{prefix}name", name))
        self._item_count += 1
        return self

    def show_timestamp(self, show: bool = True):
        self.fields.append(WidgetField(WidgetFieldType.INTEGER, "show_timestamp", str(int(show))))
        return self

    def show_column_header(self, header: ItemHistoryHeader = ItemHistoryHeader.NO):
        self.fields.append(WidgetField(WidgetFieldType.INTEGER, "show_column_header", str(header.value)))
        return self


from enum import Enum
from typing import Optional

from zbxtemplar.core.ZbxEntity import ZbxEntity, WithTags


class TriggerPriority(str, Enum):
    NOT_CLASSIFIED = "NOT_CLASSIFIED"
    INFO = "INFO"
    WARNING = "WARNING"
    AVERAGE = "AVERAGE"
    HIGH = "HIGH"
    DISASTER = "DISASTER"


class Trigger(ZbxEntity, WithTags):
    def __init__(self, name: str, expression: str,
                 priority: TriggerPriority = TriggerPriority.NOT_CLASSIFIED,
                 description: Optional[str] = None):
        super().__init__(name)
        self.expression = expression
        self.priority = priority
        self.description = description


from zbxtemplar.entities.Dashboard import Widget, WidgetField, WidgetFieldType
from zbxtemplar.entities.Graph import Graph


class ClassicGraph(Widget):
    def __init__(self, template: str, graph: Graph,
                 x: int = 0, y: int = 0, width: int = 12, height: int = 5,
                 name: str = ""):
        super().__init__(x, y, width, height, name)
        self.fields.append(
            WidgetField(WidgetFieldType.GRAPH, "graphid.0", {"host": template, "name": graph.name})
        )

    @property
    def type(self) -> str:
        return "graph"

    def widget_fields(self) -> list:
        return self.fields


from enum import Enum
from typing import List, Union

from zbxtemplar.core.ZbxEntity import ZbxEntity, WithTags
from zbxtemplar.entities.Trigger import Trigger, TriggerPriority


class ItemType(str, Enum):
    ZABBIX_PASSIVE = "ZABBIX_PASSIVE"
    TRAP = "TRAP"
    SIMPLE = "SIMPLE"
    INTERNAL = "INTERNAL"
    ZABBIX_ACTIVE = "ZABBIX_ACTIVE"
    EXTERNAL = "EXTERNAL"
    ODBC = "ODBC"
    IPMI = "IPMI"
    SSH = "SSH"
    TELNET = "TELNET"
    CALCULATED = "CALCULATED"
    JMX = "JMX"
    SNMP_TRAP = "SNMP_TRAP"
    DEPENDENT = "DEPENDENT"
    HTTP_AGENT = "HTTP_AGENT"
    SNMP_AGENT = "SNMP_AGENT"
    ITEM_TYPE_SCRIPT = "ITEM_TYPE_SCRIPT"
    ITEM_TYPE_BROWSER = "ITEM_TYPE_BROWSER"


class ValueType(str, Enum):
    FLOAT = "FLOAT"
    CHAR = "CHAR"
    LOG = "LOG"
    UNSIGNED = "UNSIGNED"
    TEXT = "TEXT"
    BINARY = "BINARY"


class Item(ZbxEntity, WithTags):
    def __init__(self, name: str, key: str, host: str,
                 type: ItemType = ItemType.ZABBIX_PASSIVE,
                 value_type: ValueType = ValueType.UNSIGNED,
                 history: str = "90d", trends: str = "365d"):
        super().__init__(name, uuid_seed=key)
        self.key = key
        self._host = host
        self.type = type
        self.value_type = value_type
        self.history = history
        self.trends = trends
        self.triggers: List[Trigger] = []

    def add_trigger(self, name: str, fn: str, op: str, threshold,
                    fn_args: tuple = (),
                    priority: TriggerPriority = TriggerPriority.NOT_CLASSIFIED,
                    description: str = None) -> 'Item':
        expression = f"{self.expr(fn, *fn_args)}{op}{threshold}"
        self.triggers.append(Trigger(name, expression, priority, description))
        return self

    def expr(self, fn: str, *args) -> str:
        base = f"{fn}(/{self._host}/{self.key})"
        if args:
            params = ",".join(str(a) for a in args)
            base = f"{fn}(/{self._host}/{self.key},{params})"
        return base


import inspect
import importlib.util
import os

import zbxtemplar.core.ZbxEntity as _zbx


class TemplarModule:
    def __init__(self):
        self.templates = []
        self.triggers = []
        self.graphs = []

    def to_export(self, version: str = "7.4") -> dict:
        groups = set()
        for t in self.templates:
            for g in t.groups:
                groups.add(g["name"])

        export = {
            "zabbix_export": {
                "version": version,
                "template_groups": [
                    {"uuid": _zbx._make_uuid(g), "name": g}
                    for g in sorted(groups)
                ],
                "templates": [t.to_dict() for t in self.templates],
            }
        }
        if self.triggers:
            export["zabbix_export"]["triggers"] = [t.to_dict() for t in self.triggers]
        if self.graphs:
            export["zabbix_export"]["graphs"] = [g.to_dict() for g in self.graphs]
        return export


def load_module(filename: str) -> dict:
    mod_name = os.path.splitext(os.path.basename(filename))[0]
    spec = importlib.util.spec_from_file_location(mod_name, filename)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    result = {}
    for name, obj in inspect.getmembers(mod, inspect.isclass):
        if issubclass(obj, TemplarModule) and obj is not TemplarModule:
            result[name] = obj()
    return result


from pathlib import Path

import yaml

from zbxtemplar.core import TemplarModule
from zbxtemplar.core.ZbxEntity import YesNo
from zbxtemplar.entities import Template, Item, Trigger, TriggerPriority, Graph, YAxisType, YAxisSide, Dashboard, \
    DashboardPage
from zbxtemplar.entities.DashboardWidget import ClassicGraph
from zbxtemplar.entities.DashboardWidget.ItemHistory import ItemHistory, ItemHistoryHeader
from zbxtemplar.entities.DashboardWidget.SimpleGraph import SimpleGraph
from zbxtemplar.entities.Item import ItemType

REFERENCE_PATH = Path(__file__).parent / "reference_template.yml"


class TestTemplate(TemplarModule):
    def __init__(self):
        super().__init__()

        template = Template(name="Test Template").add_tag("Service", "Testing")
        template.add_macro("MY_MACRO", 1, "Testing The Things")

        item1 = Item("Item 1", "item.test[1]", template.name).add_tag("Service", "Testing 1")
        item1.add_trigger(name="Simple trigger", fn="min", op=">",
                          threshold=template.get_macro("MY_MACRO"),
                          priority=TriggerPriority.HIGH, description="A single item trigger",
                          fn_args=(10,))

        item2 = Item("Item 2", "item.test[2]", template.name, type=ItemType.ZABBIX_ACTIVE).add_tag("Service", "Testing 2")
        item3 = Item(name="Item 3", key="item.test[3]", host=template.name, type=ItemType.TRAP).add_tag("Service", "Testing 3")

        template.add_item(item1).add_item(item2).add_item(item3)

        trigger_expr = (item1.expr("last") + ">" + template.get_macro("MY_MACRO")
                        + " and " + item2.expr("last") + " < " + template.get_macro("MY_MACRO"))

        graph = Graph("Test Graph", y_min=1, y_max_type=YAxisType.ITEM, y_max=item3)
        graph.add_item(item1, "1A7C11").add_item(item2, "274482", yaxisside=YAxisSide.RIGHT)

        self.templates = [template]
        self.triggers = [
            Trigger(name="Complex trigger", expression=trigger_expr,
                    priority=TriggerPriority.WARNING,
                    description="Trigger using two items"),
        ]
        self.graphs = [graph]

        graph_widget = ClassicGraph(template=template.name, graph=graph, x=0, y=0, width=36, height=5)
        first_page = DashboardPage(display_period=120)
        first_page.add_widget(graph_widget)

        item_history_widget = ItemHistory(x=36, y=0, width=18, height=5)
        item_history_widget.add_item(item2, "Item 2").add_item(item1, "Item 1")
        item_history_widget.show_timestamp(True).show_column_header(ItemHistoryHeader.HORIZONTAL)
        first_page.add_widget(item_history_widget)

        second_page = DashboardPage(name="Individual graphs")
        second_page.add_widget(SimpleGraph(name="Graph for Item1", item=item1, x=0, y=0, width=36, height=5))
        second_page.add_widget(SimpleGraph(item=item2, x=36, y=0, width=36, height=5))

        dashboard = Dashboard(name="Sample Dashboard", auto_start=YesNo.NO, display_period=60)
        dashboard.add_page(first_page).add_page(second_page)
        template.add_dashboard(dashboard)


def test_template_matches_reference():
    module = TestTemplate()
    generated = module.to_export()

    with open(REFERENCE_PATH) as f:
        expected = yaml.safe_load(f)

    assert generated == expected


from abc import ABC, abstractmethod
from enum import Enum
from typing import List

from zbxtemplar.core.ZbxEntity import ZbxEntity, YesNo


class _WidgetRefCounter:
    _counter = 0

    @classmethod
    def next(cls) -> str:
        val = cls._counter
        cls._counter += 1
        chars = []
        for _ in range(5):
            chars.append(chr(ord('A') + val % 26))
            val //= 26
        return ''.join(reversed(chars))

    @classmethod
    def reset(cls):
        cls._counter = 0


class WidgetFieldType(str, Enum):
    INTEGER = "INTEGER"
    STRING = "STRING"
    ITEM = "ITEM"
    ITEM_PROTOTYPE = "ITEM_PROTOTYPE"
    GRAPH = "GRAPH"
    GRAPH_PROTOTYPE = "GRAPH_PROTOTYPE"
    MAP = "MAP"
    SERVICE = "SERVICE"
    SLA = "SLA"
    USER = "USER"
    ACTION = "ACTION"
    MEDIA_TYPE = "MEDIA_TYPE"


class WidgetField:
    def __init__(self, type: WidgetFieldType, name: str, value):
        self.type = type
        self.name = name
        self.value = value

    def to_dict(self):
        value = str(self.value) if isinstance(self.value, (int, float)) else self.value
        return {"type": self.type.value, "name": self.name, "value": value}


class Widget(ABC):
    def __init__(self, x: int, y: int, width: int, height: int, name: str = ""):
        self.name = name
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.fields: List[WidgetField] = [
            WidgetField(WidgetFieldType.STRING, "reference", _WidgetRefCounter.next())
        ]

    @property
    @abstractmethod
    def type(self) -> str:
        pass

    @abstractmethod
    def widget_fields(self) -> list:
        pass

    def to_dict(self):
        result = {"type": self.type}
        if self.name:
            result["name"] = self.name
        if self.x:
            result["x"] = str(self.x)
        if self.y:
            result["y"] = str(self.y)
        result["width"] = str(self.width)
        result["height"] = str(self.height)
        result["fields"] = [f.to_dict() for f in self.fields]
        return result


class DashboardPage:
    def __init__(self, name: str = "", display_period: int = 0):
        self.name = name
        self.display_period = display_period
        self.widgets: List[Widget] = []

    def add_widget(self, widget: Widget):
        self.widgets.append(widget)

    def to_dict(self):
        result = {}
        if self.name:
            result["name"] = self.name
        if self.display_period:
            result["display_period"] = str(self.display_period)
        if self.widgets:
            result["widgets"] = [w.to_dict() for w in self.widgets]
        return result


class Dashboard(ZbxEntity):
    def __init__(self, name: str, display_period: int = 0,
                 auto_start: YesNo = YesNo.YES):
        super().__init__(name)
        self.display_period = str(display_period)
        self.auto_start = auto_start
        self.pages: List[DashboardPage] = []

    def add_page(self, page: DashboardPage) -> 'Dashboard':
        self.pages.append(page)
        return self


from enum import Enum
from typing import List, Union

from zbxtemplar.core.ZbxEntity import ZbxEntity, YesNo
from zbxtemplar.entities import Item


class GraphType(str, Enum):
    NORMAL = "NORMAL"
    STACKED = "STACKED"
    PIE = "PIE"
    EXPLODED = "EXPLODED"


class YAxisType(str, Enum):
    CALCULATED = "CALCULATED"
    FIXED = "FIXED"
    ITEM = "ITEM"


class DrawType(str, Enum):
    SINGLE_LINE = "SINGLE_LINE"
    FILLED_REGION = "FILLED_REGION"
    BOLD_LINE = "BOLD_LINE"
    DOTTED_LINE = "DOTTED_LINE"
    DASHED_LINE = "DASHED_LINE"
    GRADIENT_LINE = "GRADIENT_LINE"


class CalcFnc(str, Enum):
    MIN = "MIN"
    AVG = "AVG"
    MAX = "MAX"
    ALL = "ALL"
    LAST = "LAST"


class GraphItemType(str, Enum):
    SIMPLE = "SIMPLE"
    GRAPH_SUM = "GRAPH_SUM"


class YAxisSide(str, Enum):
    LEFT = "LEFT"
    RIGHT = "RIGHT"


class GraphItem:
    def __init__(self, item: Item, color: str, order: int = 0,
                 drawtype: DrawType = DrawType.SINGLE_LINE,
                 calc_fnc: CalcFnc = CalcFnc.AVG,
                 type: GraphItemType = GraphItemType.SIMPLE,
                 yaxisside: YAxisSide = YAxisSide.LEFT):
        self.item = {"host": item._host, "key": item.key}
        self.color = color
        self.sortorder = str(order)
        self.drawtype = drawtype
        self.calc_fnc = calc_fnc
        self.type = type
        self.yaxisside = yaxisside


class Graph(ZbxEntity):
    def __init__(self, name: str, type: GraphType = GraphType.NORMAL,
                 show_triggers: YesNo = YesNo.YES,
                 show_legend: YesNo = YesNo.YES,
                 y_min_type: YAxisType = YAxisType.CALCULATED,
                 y_max_type: YAxisType = YAxisType.CALCULATED,
                 y_min: Union[int, float, Item, str, None] = None,
                 y_max: Union[int, float, Item, str, None] = None):
        super().__init__(name)
        self.type = type
        self.show_triggers = show_triggers
        self.show_legend = show_legend
        self.ymin_type_1 = y_min_type
        self.ymax_type_1 = y_max_type
        self._resolve_y("min", y_min, y_min_type)
        self._resolve_y("max", y_max, y_max_type)
        self.graph_items: List[GraphItem] = []

    def _resolve_y(self, axis: str, value, axis_type: YAxisType):
        if axis_type == YAxisType.ITEM:
            if not isinstance(value, Item):
                raise ValueError(f"y_{axis}: type=ITEM requires an Item value")
            setattr(self, f"y{axis}_item_1", {"host": value._host, "key": value.key})
        elif value is not None:
            if isinstance(value, Item):
                raise ValueError(f"y_{axis}: Item value requires y_{axis}_type=ITEM")
            setattr(self, f"yaxis{axis}", str(value))

    def add_item(self, item: Item, color: str, order: int = -1,
                 drawtype: DrawType = DrawType.SINGLE_LINE,
                 calc_fnc: CalcFnc = CalcFnc.AVG,
                 type: GraphItemType = GraphItemType.SIMPLE,
                 yaxisside: YAxisSide = YAxisSide.LEFT) -> 'Graph':
        if order == -1:
            order = len(self.graph_items)
        self.graph_items.append(GraphItem(item, color, order, drawtype, calc_fnc, type, yaxisside))
        return self


from zbxtemplar.entities import Item
from zbxtemplar.entities.Dashboard import Widget, WidgetField, WidgetFieldType


class SimpleGraph(Widget):
    def __init__(self, item: Item,
                 x: int = 0, y: int = 0, width: int = 12, height: int = 5,
                 name: str = ""):
        super().__init__(x, y, width, height, name)
        self.fields.append(WidgetField(WidgetFieldType.INTEGER, "source_type", 1))
        self.fields.append(
            WidgetField(WidgetFieldType.ITEM, "itemid.0", {"host": item._host, "key": item.key})
        )

    @property
    def type(self) -> str:
        return "graph"

    def widget_fields(self) -> list:
        return self.fields


from typing import Any, List, Optional, Dict, Union
from enum import Enum
import uuid

class YesNo(str, Enum):
    NO = "NO"
    YES = "YES"

ZBX_TEMPLAR_NAMESPACE = "Zbx Templar"
ZBX_TEMPLAR_TEMPLATE_GROUP = "Templar Templates"
_NAMESPACE_UUID = uuid.uuid5(uuid.NAMESPACE_DNS, ZBX_TEMPLAR_NAMESPACE)

def set_uuid_namespace(namespace: str):
    global ZBX_TEMPLAR_NAMESPACE, _NAMESPACE_UUID
    ZBX_TEMPLAR_NAMESPACE = namespace
    _NAMESPACE_UUID = uuid.uuid5(uuid.NAMESPACE_DNS, namespace)

def set_template_group(group: str):
    global ZBX_TEMPLAR_TEMPLATE_GROUP
    ZBX_TEMPLAR_TEMPLATE_GROUP = group

def _make_uuid(seed: str) -> str:
    """Deterministic UUID from seed, masked as UUIDv4. Zabbix rejects UUID5 on import."""
    h = uuid.uuid5(_NAMESPACE_UUID, seed).hex
    return h[:12] + '4' + h[13:16] + hex(0x8 | (int(h[16], 16) & 0x3))[2:] + h[17:]

def _serialize(value):
    if hasattr(value, 'to_dict'):
        return value.to_dict()
    if isinstance(value, dict):
        return {k: _serialize(v) for k, v in value.items()}
    if isinstance(value, list):
        return [_serialize(v) for v in value]
    if isinstance(value, Enum):
        return value.value
    if hasattr(value, '__dict__') and not isinstance(value, (str, int, float, bool)):
        return {k: _serialize(v) for k, v in value.__dict__.items()}
    return value

class ZbxEntity:
    def __init__(self, name: str, uuid_seed: str = None):
        super().__init__()
        self.name = name
        self.uuid = _make_uuid(uuid_seed or name)

    def to_dict(self) -> dict[str, Any]:
        # Attributes starting with _ are internal-only (e.g. _host) and skipped here
        result = {}
        for key, value in self.__dict__.items():
            if key.startswith('_'):
                continue
            if value is None or value == {} or value == []:
                continue
            if key == 'tags' and hasattr(self, 'tags_to_list'):
                result[key] = self.tags_to_list()
            elif key == 'macros' and hasattr(self, 'macros_to_list'):
                result[key] = self.macros_to_list()
            else:
                result[key] = _serialize(value)
        return result

class Tag:
    def __init__(self, name: str, value: str = ""):
        self.name = name
        self.value = value

    def to_dict(self):
        return {"tag": self.name, "value": self.value}

class WithTags():
    def __init__(self):
        super().__init__()
        self.tags: Dict[str, Tag] = {}

    def add_tag(self, tag: str, value: str = ""):
        self.tags[tag] = Tag(name=tag, value=value)
        return self

    def tags_to_list(self):
        return [t.to_dict() for t in self.tags.values()]

    def load_tags(self, tags: Dict[str, str]):
        """Load tags from a dict.

        Args:
            tags: Mapping of tag name to value.
                Example: ``{"Application": "CPU", "Component": "OS"}``
        """
        for tag, value in tags.items():
            self.add_tag(tag, value)
        return self

class Macro():
    def __init__(self, name: str, value: str, description: Optional[str] = None):
        self.name = name
        self.value = value
        self.description = description

    @property
    def full_name(self) -> str:
        return f"{{${self.name}}}"

    def __str__(self):
        return self.full_name

    def __add__(self, other):
        return str(self) + other

    def __radd__(self, other):
        return other + str(self)

    def to_dict(self):
        result = {"macro": self.full_name, "value": self.value}
        if self.description is not None:
            result["description"] = self.description
        return result

class WithMacros():
    def __init__(self):
        super().__init__()
        self.macros: Dict[str, Macro] = {}

    def add_macro(self, name: str, value: Union[str, int], description: Union[str, None] = None):
        clean_name = name.replace("{$", "").replace("}", "")
        self.macros[clean_name] = Macro(name=clean_name, value=str(value), description=description)
        return self

    def load_macros(self, macros: Union[Dict[str, Union[str, tuple]], List[dict]]):
        """Load macros from a dict or a list of dicts.

        Accepts two formats:

        **Dict** — for programmatic use::

            load_macros({"TIMEOUT": "30"})
            load_macros({"TIMEOUT": ("30", "Connection timeout")})

        **List of dicts** — for data loaded from YAML/XML::

            load_macros([
                {"name": "TIMEOUT", "value": "30", "description": "Connection timeout"}
            ])

        Args:
            macros: Either a ``{name: value}`` / ``{name: (value, description)}`` dict,
                or a list of dicts with ``name``, ``value``, and optional ``description`` keys.
        """
        if isinstance(macros, dict):
            for name, value in macros.items():
                if isinstance(value, tuple):
                    self.add_macro(name, value[0], value[1] if len(value) > 1 else None)
                else:
                    self.add_macro(name, value)
        else:
            for entry in macros:
                self.add_macro(entry["name"], entry["value"], entry.get("description"))
        return self

    def macros_to_list(self):
        return [m.to_dict() for m in self.macros.values()]

    def get_macro(self, name: str) -> Optional[Macro]:
        clean_name = name.replace("{$", "").replace("}", "")
        return self.macros.get(clean_name)




from zbxtemplar.entities.Item import Item
from zbxtemplar.entities.Template import Template
from zbxtemplar.entities.Trigger import Trigger, TriggerPriority
from zbxtemplar.entities.Graph import Graph, GraphType, YAxisType, DrawType, CalcFnc, GraphItemType, YAxisSide
from zbxtemplar.entities.Dashboard import Dashboard, DashboardPage
from zbxtemplar.core.ZbxEntity import YesNo


from typing import List, Union

import zbxtemplar.core.ZbxEntity as _zbx
from zbxtemplar.core.ZbxEntity import ZbxEntity, WithTags, WithMacros
from zbxtemplar.entities.Dashboard import Widget, Dashboard
from zbxtemplar.entities.Item import Item


class Template(ZbxEntity, WithTags, WithMacros):
    def __init__(self, name: str, groups: Union[None, List[str]] = None):
        super().__init__(name)
        if groups is None:
            groups = [_zbx.ZBX_TEMPLAR_TEMPLATE_GROUP]
        self.template = name
        self.items: List[Item] = []
        self.dashboards: List[Dashboard] = []
        self.groups = [{"name": g} for g in groups]

    def add_item(self, item: Item):
        item._host = self.name
        self.items.append(item)
        return self

    def add_dashboard(self, dashboard: Dashboard):
        self.dashboards.append(dashboard)
        return self




import argparse
import yaml

from zbxtemplar.core.ZbxEntity import set_uuid_namespace, set_template_group
from zbxtemplar.core.TemplarModule import load_module


def main():
    parser = argparse.ArgumentParser(description="Zbx Templar — Zabbix template generator")
    parser.add_argument("module", help="Path to a Python module (.py) with TemplarModule subclass(es)")
    parser.add_argument("output", help="Output YAML file path")
    parser.add_argument("--namespace", help="UUID namespace for deterministic ID generation")
    parser.add_argument("--template-group", help="Default template group name")
    args = parser.parse_args()

    if args.namespace:
        set_uuid_namespace(args.namespace)
    if args.template_group:
        set_template_group(args.template_group)

    modules = load_module(args.module)
    if not modules:
        print(f"No TemplarModule subclasses found in {args.module}")
        return 1

    for name, mod in modules.items():
        export = mod.to_export()
        with open(args.output, "w") as f:
            yaml.dump(export, f, default_flow_style=False, sort_keys=False)
        print(f"{name} -> {args.output}")

    return 0


if __name__ == "__main__":
    raise SystemExit(main())